@@ -64,6 +64,9 @@ extern int errno;
 #endif
 
 #include "regex.h"
+#include "timespec.h"
+#include "buildcmd.h"
+
 
 #ifndef S_IFLNK
 #define lstat stat
@@ -85,6 +88,7 @@ int optionp_stat PARAMS((const char *name, struct stat *p));
 int optionh_stat PARAMS((const char *name, struct stat *p));
 int debug_stat   PARAMS((const char *file, struct stat *bufp));
 
+void set_stat_placeholders PARAMS((struct stat *p));
 int get_statinfo PARAMS((const char *pathname, const char *name, struct stat *p));
 
 #if ! defined HAVE_FCHDIR && ! defined fchdir
@@ -220,26 +224,22 @@ struct size_val
   uintmax_t size;
 };
 
-#define NEW_EXEC 1
-/*
-#undef NEW_EXEC 
-*/
 
-#if !defined(NEW_EXEC)
-struct path_arg
+enum xval 
+  {
+    XVAL_ATIME, XVAL_BIRTHTIME, XVAL_CTIME, XVAL_MTIME, XVAL_TIME
+  };
+
+struct time_val
 {
-  short offset;			/* Offset in `vec' of this arg. */
-  short count;			/* Number of path replacements in this arg. */
-  char *origarg;		/* Arg with "{}" intact. */
+  enum xval            xval; 
+  enum comparison_type kind;
+  struct timespec      ts;
 };
-#endif
-
-#include "buildcmd.h"
 
+    
 struct exec_val
 {
-#if defined(NEW_EXEC)
-  /* new-style */
   boolean multiple;		/* -exec {} \+ denotes multiple argument. */
   struct buildcmd_control ctl;
   struct buildcmd_state   state;
@@ -247,10 +247,7 @@ struct exec_val
   int num_args;
   boolean use_current_dir;      /* If nonzero, don't chdir to start dir */
   boolean close_stdin;		/* If true, close stdin in the child. */
-#else
-  struct path_arg *paths;	/* Array of args with path replacements. */
-  char **vec;			/* Array of args to pass to program. */
-#endif
+  int dirfd;			/* The directory to do the exec in. */
 };
 
 /* The format string for a -printf or -fprintf is chopped into one or
@@ -259,12 +256,17 @@ struct exec_val
    each \c and `%' conversion is a segment. */
 
 /* Special values for the `kind' field of `struct segment'. */
-#define KIND_PLAIN 0		/* Segment containing just plain text. */
-#define KIND_STOP 1		/* \c -- stop printing and flush output. */
+enum SegmentKind 
+  {
+    KIND_PLAIN=0,		/* Segment containing just plain text. */
+    KIND_STOP=1,		/* \c -- stop printing and flush output. */
+    KIND_FORMAT,		/* Regular format */
+  };
 
 struct segment
 {
-  int kind;			/* Format chars or KIND_{PLAIN,STOP}. */
+  enum SegmentKind segkind;     /* KIND_FORMAT, KIND_PLAIN, KIND_STOP */
+  char format_char[2];		/* Format chars if kind is KIND_FORMAT */
   char *text;			/* Plain text or `%' format string. */
   int text_len;			/* Length of `text'. */
   struct segment *next;		/* Next segment for this predicate. */
@@ -346,12 +348,11 @@ struct predicate
     char *str;			/* fstype [i]lname [i]name [i]path */
     struct re_pattern_buffer *regex; /* regex */
     struct exec_val exec_vec;	/* exec ok */
-    struct long_val info;	/* atime ctime gid inum links mtime
-                                   size uid */
+    struct long_val numinfo;	/* gid inum links  uid */
     struct size_val size;	/* size */
     uid_t uid;			/* user */
     gid_t gid;			/* group */
-    time_t time;		/* newer */
+    struct time_val reftime;	/* newer newerXY anewer cnewer mtime atime ctime mmin amin cmin */
     struct perm_val perm;	/* perm */
     struct dir_id   fileid;	/* samefile */
     mode_t type;		/* type */
@@ -374,9 +375,9 @@ struct predicate
 };
 
 /* find.c, ftsfind.c */
-boolean is_fts_enabled();
-
-
+boolean is_fts_enabled(int *ftsoptions);
+int get_start_dirfd(void);
+int get_current_dirfd(void);
 
 /* find library function declarations.  */
 
@@ -386,9 +387,6 @@ char *dirname PARAMS((char *path));
 /* error.c */
 void error PARAMS((int status, int errnum, char *message, ...));
 
-/* listfile.c */
-char *get_link_name PARAMS((char *name, char *relname));
-
 /* stpcpy.c */
 #if !HAVE_STPCPY
 char *stpcpy PARAMS((char *dest, const char *src));
@@ -434,6 +432,7 @@ enum arg_type
     ARG_NOOP,			/* does nothing, returns true, internal use only */
     ARG_POSITIONAL_OPTION,	/* options whose position is important (-follow) */
     ARG_TEST,			/* a like -name */
+    ARG_SPECIAL_PARSE,		/* complex to parse, don't eat the test name before calling parse_xx(). */
     ARG_PUNCTUATION,		/* like -o or ( */
     ARG_ACTION			/* like -print */
   };
@@ -497,6 +496,7 @@ boolean pred_mtime PARAMS((char *pathname, struct stat *stat_buf, struct predica
 boolean pred_name PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_negate PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_newer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_newerXY PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_nogroup PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_nouser PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_ok PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -550,13 +550,14 @@ struct predicate *insert_primary_withpred PARAMS((const struct parser_table *ent
 void usage PARAMS((FILE *fp, int status, char *msg));
 extern boolean check_nofollow(void);
 void complete_pending_execs(struct predicate *p);
-void complete_pending_execdirs(struct predicate *p);
+void complete_pending_execdirs(int dirfd); /* Passing dirfd is an unpleasant CodeSmell. */
+
 int process_leading_options PARAMS((int argc, char *argv[]));
 void set_option_defaults PARAMS((struct options *p));
 
 
 /* find.c. */
-int get_info PARAMS((const char *pathname, const char *name, struct stat *p, struct predicate *pred_ptr));
+int get_info PARAMS((const char *pathname, struct stat *p, struct predicate *pred_ptr));
 int following_links PARAMS((void));
 int digest_mode PARAMS((mode_t mode, const char *pathname, const char *name, struct stat *pstat, boolean leaf));
 boolean default_prints PARAMS((struct predicate *pred));
@@ -570,7 +571,8 @@ enum DebugOption
     DebugStat             = 2,
     DebugSearch           = 4,
     DebugTreeOpt          = 8,
-    DebugHelp             = 16
+    DebugHelp             = 16,
+    DebugExec             = 32
   };
 
 struct options
@@ -605,7 +607,7 @@ struct options
    */
   boolean warnings;
   
-  time_t start_time;		/* Time at start of execution.  */
+  struct timespec      start_time;		/* Time at start of execution.  */
   
   /* Seconds between 00:00 1/1/70 and either one day before now
      (the default), or the start of today (if -daystart is given). */
@@ -659,6 +661,10 @@ struct state
   /* The file being operated on, relative to the current directory.
      Used for stat, readlink, remove, and opendir.  */
   char *rel_pathname;
+  /* The directory fd to which rel_pathname is relative.  Thsi is relevant
+   * when we're navigating the hierarchy with fts() and using FTS_CWDFD.
+   */
+  int cwd_dir_fd;
 
   /* Length of starting path. */
   int starting_path_length;
@@ -669,6 +675,12 @@ struct state
   
   /* Status value to return to system. */
   int exit_status;
+
+  /* True if there are any execdirs.  This saves us a pair of fchdir()
+   * calls for every directory we leave if it is false.  This is just
+   * an optimisation.  Set to true if you want to be conservative.
+   */
+  boolean execdirs_outstanding;
 };
 
 /* finddata.c */
