@@ -444,9 +444,32 @@ consider_visiting(FTS *p, FTSENT *ent)
 	  return;
 	}
     }
+  else if (ent->fts_info == FTS_NS)
+    {
+      if (ent->fts_level == 0)
+	{
+	  /* e.g., nonexistent starting point */
+	  error(0, ent->fts_errno, ent->fts_path);
+	  error_severity(1);	/* remember problem */
+	  return;
+	}
+      else
+	{
+	  /* The following if statement fixes Savannah bug #19605
+	   * (failure to diagnose a symbolic link loop)
+	   */
+	  if (symlink_loop(ent->fts_accpath))
+	    {
+	      error(0, ELOOP, ent->fts_path);
+	      error_severity(1);
+	      return;
+	    }
+	}
+    }
   
-  /* Not an error, cope with the usual cases. */
-  if (ent->fts_info == FTS_NSOK)
+  /* Cope with the usual cases. */
+  if (ent->fts_info == FTS_NSOK
+      || ent->fts_info == FTS_NS /* e.g. symlink loop */)
     {
       assert(!state.have_stat);
       assert(!state.have_type);
@@ -717,6 +740,6 @@ boolean
 is_fts_enabled(int *fts_options)
 {
   /* this version of find (i.e. this main()) uses fts. */
-  *fts_options = fts_options;
+  *fts_options = ftsoptions;
   return true;
 }
