@@ -309,6 +309,7 @@ predlist_insert(struct predlist *list,
     list->tail = list->head;
 }
 
+#if 0
 static void 
 predlist_dump(FILE *fp, const char *label, const struct predlist *list)
 {
@@ -334,6 +335,7 @@ predlist_merge_nosort(struct predlist *list,
     }
 }
 
+#endif
 static int
 pred_cost_compare(const struct predicate *p1, const struct predicate *p2, boolean wantfailure)
 {
@@ -1017,6 +1019,7 @@ static struct pred_cost_lookup costlookup[] =
     { pred_name	     ,  NeedsNothing         },
     { pred_negate    ,  NeedsNothing,        },
     { pred_newer     ,  NeedsStatInfo,       },
+    { pred_newerXY   ,  NeedsStatInfo,       },
     { pred_nogroup   ,  NeedsStatInfo        }, /* true for amortised cost if caching is on */
     { pred_nouser    ,  NeedsStatInfo        }, /* true for amortised cost if caching is on */
     { pred_ok        ,  NeedsUserInteraction },
@@ -1181,15 +1184,13 @@ getrate(const struct predicate *p)
   if (p)
     return p->est_success_rate;
   else
-    return 1.0;
+    return 1.0f;
 }
 
 
 float 
 calculate_derived_rates(struct predicate *p)
 {
-  float rate;
-
   assert(NULL != p);
 
   if (p->pred_right)
@@ -1199,48 +1200,57 @@ calculate_derived_rates(struct predicate *p)
 
   assert(p->p_type != CLOSE_PAREN);
   assert(p->p_type != OPEN_PAREN);
-  
-  if (p->p_type == PRIMARY_TYPE)
+
+  switch (p->p_type)
     {
+    case NO_TYPE:
       assert(NULL == p->pred_right);
       assert(NULL == p->pred_left);
       return p->est_success_rate;
-    }
-  else if (p->p_type == UNI_OP)
-    {
+      
+    case PRIMARY_TYPE:
+      assert(NULL == p->pred_right);
+      assert(NULL == p->pred_left);
+      return p->est_success_rate;
+
+    case UNI_OP:
       /* Unary operators must have exactly one operand */
       assert(pred_negate == p->pred_func);
       assert(NULL == p->pred_left);
-      rate = 1.0 - p->pred_right->est_success_rate;
-    }
-  else if (p->p_type == BI_OP)
-    {
-      /* Binary operators must have two operands */
-      if (pred_and == p->pred_func)
-	{
-	  rate = getrate(p->pred_right) * getrate(p->pred_left);
-	}
-      else if (pred_comma == p->pred_func)
-	{
-	  rate = 1.0;
-	}
-      else if (pred_or == p->pred_func)
-	{
-	  rate = getrate(p->pred_right) + getrate(p->pred_left);
-	}
-    }
-  else if (p->p_type == CLOSE_PAREN || p->p_type == OPEN_PAREN)	    
-    {
-      rate = 1.0;
-    }
-  else if (p->p_type == NO_TYPE)
-    {
-      assert(NULL == p->pred_right);
-      assert(NULL == p->pred_left);
+      p->est_success_rate = (1.0 - p->pred_right->est_success_rate);
+      return p->est_success_rate;
+
+    case BI_OP:
+      {
+	float rate;
+	/* Binary operators must have two operands */
+	if (pred_and == p->pred_func)
+	  {
+	    rate = getrate(p->pred_right) * getrate(p->pred_left);
+	  }
+	else if (pred_comma == p->pred_func)
+	  {
+	    rate = 1.0f;
+	  }
+	else if (pred_or == p->pred_func)
+	  {
+	    rate = getrate(p->pred_right) + getrate(p->pred_left);
+	  }
+	else
+	  {
+	    /* only and, or and comma are BI_OP. */
+	    assert(0);
+	    rate = 0.0f;
+	  }
+	p->est_success_rate = constrain_rate(rate);
+      }
+      return p->est_success_rate;
+
+    case OPEN_PAREN:
+    case CLOSE_PAREN:
+      p->est_success_rate = 1.0;
       return p->est_success_rate;
     }
-  p->est_success_rate = constrain_rate(rate);
-  return p->est_success_rate;
 }
 
 /* opt_expr() rearranges predicates such that each left subtree is
@@ -1277,14 +1287,14 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
 
   predicates = NULL;
   
-  /* Find where in ARGV the predicates begin by skipping the list of start points. */
+  /* Find where in ARGV the predicates begin by skipping the list of
+   * start points.
+   */
   for (i = end_of_leading_options; i < argc && !looks_like_expression(argv[i], true); i++)
     {
-      /* fprintf(stderr, "Looks like %s is not a predicate\n", argv[i]); */
       /* Do nothing. */ ;
     }
   
-  /* XXX: beginning of bit we need factor out of both find.c and ftsfind.c */
   /* Enclose the expression in `( ... )' so a default -print will
      apply to the whole expression. */
   entry_open  = find_parser("(");
@@ -1314,19 +1324,40 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
       if (parse_entry == NULL)
 	{
 	  /* Command line option not recognized */
-	  error (1, 0, _("invalid predicate `%s'"), predicate_name);
+	  error (1, 0, _("unknown predicate `%s'"), predicate_name);
+	}
+
+      /* We have recognised a test of the form -foo.  Eat that, 
+       * unless it is a predicate like -newerXY.
+       */
+      if (parse_entry->type != ARG_SPECIAL_PARSE)
+	{
+	  i++;
 	}
-      
-      i++;
       oldi = i;
       if (!(*(parse_entry->parser_func)) (parse_entry, argv, &i))
 	{
-	  if (argv[i] == NULL)
-	    /* Command line option requires an argument */
-	    error (1, 0, _("missing argument to `%s'"), predicate_name);
+	  if (argv[i])
+	    {
+	      if ( (ARG_SPECIAL_PARSE == parse_entry->type) && (i == oldi) )
+		{
+		  /* The special parse function spat out the
+		   * predicate.  It must be invalid, or not tasty.
+		   */
+		  error (1, 0, _("invalid predicate `%s'"),
+			 argv[i], predicate_name);
+		}
+	      else
+		{
+		  error (1, 0, _("invalid argument `%s' to `%s'"),
+			 argv[i], predicate_name);
+		}
+	    }
 	  else
-	    error (1, 0, _("invalid argument `%s' to `%s'"),
-		   argv[i], predicate_name);
+	    {
+	      /* Command line option requires an argument */
+	      error (1, 0, _("missing argument to `%s'"), predicate_name);
+	    }
 	}
       else
 	{
