@@ -21,8 +21,10 @@
 #if !defined LISTFILE_H
 # define LISTFILE_H
 
-void list_file (char *name, char *relname, struct stat *statp, time_t current_time, int output_block_size, int literal_control_chars, FILE *stream);
+void list_file (char *name, int dir_fd, char *relname, const struct stat *statp, time_t current_time, int output_block_size, int literal_control_chars, FILE *stream);
 
+char * get_link_name_at (const char *name, int dirfd, char *relname);
 
+size_t file_blocksize(const struct stat *p);
 
 #endif
