@@ -1,6 +1,6 @@
 /* locate -- search databases for filenames that match patterns
    Copyright (C) 1994, 1996, 1998, 1999, 2000, 2003,
-                 2004, 2005 Free Software Foundation, Inc.
+                 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -65,6 +65,7 @@
 #include <signal.h>
 #include <ctype.h>
 #include <sys/types.h>
+#include <grp.h>		/* for setgroups() */
 #include <sys/stat.h>
 #include <time.h>
 #include <fnmatch.h>
@@ -84,13 +85,8 @@
 
 #define NDEBUG
 #include <assert.h>
-
-#if defined(HAVE_STRING_H) || defined(STDC_HEADERS)
 #include <string.h>
-#else
-#include <strings.h>
-#define strchr index
-#endif
+
 
 #ifdef STDC_HEADERS
 #include <stdlib.h>
@@ -140,9 +136,9 @@ extern int errno;
 #include "quotearg.h"
 #include "printquoted.h"
 #include "regextype.h"
+#include "gnulib-version.h"
 
-
-/* Note that this evaluates C many times.  */
+/* Note that this evaluates Ch many times.  */
 #ifdef _LIBC
 # define TOUPPER(Ch) toupper (Ch)
 # define TOLOWER(Ch) tolower (Ch)
@@ -155,13 +151,12 @@ extern int errno;
 
 /* Warn if a database is older than this.  8 days allows for a weekly
    update that takes up to a day to perform.  */
-#define WARN_NUMBER_UNITS (8)
+static unsigned int warn_number_units = 8;
+
 /* Printable name of units used in WARN_SECONDS */
 static const char warn_name_units[] = N_("days");
 #define SECONDS_PER_UNIT (60 * 60 * 24)
 
-#define WARN_SECONDS ((SECONDS_PER_UNIT) * (WARN_NUMBER_UNITS))
-
 enum visit_result
   {
     VISIT_CONTINUE = 1,  /* please call the next visitor */
@@ -190,11 +185,58 @@ static bool stdout_is_a_tty;
 static bool print_quoted_filename;
 static bool results_were_filtered;
 
-static char*  slocate_db_pathname = "/var/lib/slocate/slocate.db";
+/* static char*  slocate_db_pathname = "/var/lib/slocate/slocate.db"; */
 
 static const char *selected_secure_db = NULL;
 
 
+/* Change the number of days old the database can be 
+ * before we complain about it. 
+ */
+static void
+set_max_db_age(const char *s) 
+{
+  char *end;
+  unsigned long int val;
+  /* XXX: we ignore the case where the input is negative, which is allowed(!). */
+
+  if (0 == *s)
+    {
+      error(1, 0,
+	    _("The argument argument for option --max-database-age must not be empty"));
+    }
+  
+  
+  /* We have to set errno here, otherwise when the function returns ULONG_MAX,
+   * we would not be able to tell if that is the correct answer, or whether it
+   * signifies an error.
+   */
+  errno = 0;
+  val = strtoul(s, &end, 10);
+  
+  /* Diagnose number too large, non-numbes and trailing junk. */
+  if ((ULONG_MAX == val && ERANGE == errno) ||
+      (0 == val && EINVAL == errno))
+    {
+      error(1, errno,
+	    _("Invalid argument `%s' for option --max-database-age"),
+	    s);
+    }
+  else if (*end)
+    {
+      /* errno wasn't set, don't print its message */
+      error(1, 0,
+	    _("Invalid argument `%s' for option --max-database-age"),
+	    s);
+    }
+  else
+    {
+      warn_number_units = val;
+    }
+}
+
+
+
 /* Read in a 16-bit int, high byte first (network byte order).  */
 
 static short
@@ -269,17 +311,6 @@ locate_read_str(char **buf, size_t *siz, FILE *fp, int delimiter, int offs)
 }
 
 
-static void
-lc_strcpy(char *dest, const char *src)
-{
-  while (*src)
-    {
-      *dest++ = TOLOWER(*src);
-      ++src;
-    }
-  *dest = 0;
-}
-
 struct locate_limits
 {
   uintmax_t limit;
@@ -309,12 +340,6 @@ struct stringbuf
 static struct stringbuf casebuf;
 
 
-struct casefolder
-{
-  const char *pattern;
-  struct stringbuf *pbuf;
-};
-
 struct regular_expression
 {
   struct re_pattern_buffer regex; /* for --regex */
@@ -599,21 +624,6 @@ visit_basename(struct process_data *procdata, void *context)
 }
 
 
-static int
-visit_casefold(struct process_data *procdata, void *context)
-{
-  struct stringbuf *b = context;
-
-  if (*b->preqlen+1 > b->buffersize)
-    {
-      b->buffer = xrealloc(b->buffer, *b->preqlen+1); /* XXX: consider using extendbuf(). */
-      b->buffersize = *b->preqlen+1;
-    }
-  lc_strcpy(b->buffer, procdata->munged_filename);
-
-  return VISIT_CONTINUE;
-}
-  
 /* visit_existing_follow implements -L -e */
 static int
 visit_existing_follow(struct process_data *procdata, void *context)
@@ -707,7 +717,7 @@ visit_substring_match_nocasefold(struct process_data *procdata, void *context)
 {
   const char *pattern = context;
 
-  if (NULL != strstr(procdata->munged_filename, pattern))
+  if (NULL != mbsstr(procdata->munged_filename, pattern))
     return VISIT_ACCEPTED;
   else
     return VISIT_REJECTED;
@@ -716,11 +726,9 @@ visit_substring_match_nocasefold(struct process_data *procdata, void *context)
 static int
 visit_substring_match_casefold(struct process_data *procdata, void *context)
 {
-  const struct casefolder * p = context;
-  const struct stringbuf * b = p->pbuf;
-  (void) procdata;
+  const char *pattern = context;
 
-  if (NULL != strstr(b->buffer, p->pattern))
+  if (NULL != mbscasestr(procdata->munged_filename, pattern))
     return VISIT_ACCEPTED;
   else
     return VISIT_REJECTED;
@@ -910,8 +918,6 @@ looking_at_slocate_locatedb (const char *filename,
 			     size_t len,
 			     int *seclevel)
 {
-  char slocate_magic[] = "1";
-  size_t lenwanted = sizeof(slocate_magic);
   assert(len <= 2);
   
   if (len < 2)
@@ -946,6 +952,11 @@ looking_at_slocate_locatedb (const char *filename,
 		  return 1;
 		}
 	    }
+	  else
+	    {
+	      /* Not a digit. */
+	      return 0;
+	    }
 	}
       else
 	{
@@ -976,7 +987,6 @@ search_one_database (int argc,
 {
   char *pathpart; 		/* A pattern to consider. */
   int argn;			/* Index to current pattern in argv. */
-  int need_fold;	/* Set when folding and any pattern is non-glob. */
   int nread;		     /* number of bytes read from an entry. */
   struct process_data procdata;	/* Storage for data shared with visitors. */
   int slocate_seclevel;
@@ -1016,6 +1026,7 @@ search_one_database (int argc,
 
   nread = fread (procdata.original_filename, 1, SLOCATE_DB_MAGIC_LEN,
 		 procdata.fp);
+  slocate_seclevel = 0;
   if (looking_at_slocate_locatedb(procdata.dbfile,
 				  procdata.original_filename,
 				  nread,
@@ -1110,24 +1121,6 @@ search_one_database (int argc,
   if (basename_only)
     add_visitor(visit_basename, NULL);
   
-  /* See if we need fold. */
-  if (ignore_case && !regex)
-    for ( argn = 0; argn < argc; argn++ )
-      {
-        pathpart = argv[argn];
-        if (!contains_metacharacter(pathpart))
-	  {
-	    need_fold = 1;
-	    break;
-	  }
-      }
-
-  if (need_fold)
-    {
-      add_visitor(visit_casefold, &casebuf);
-      casebuf.preqlen = &procdata.pathsize;
-    }
-  
   /* Add an inspector for each pattern we're looking for. */
   for ( argn = 0; argn < argc; argn++ )
     {
@@ -1174,20 +1167,9 @@ search_one_database (int argc,
 	   * James Youngman <jay@gnu.org> 
 	   */
 	  if (ignore_case)
-	    {
-	      struct casefolder * cf = xmalloc(sizeof(*cf));
-	      cf->pattern = pathpart;
-	      cf->pbuf = &casebuf;
-	      add_visitor(visit_substring_match_casefold, cf);
-	      /* If we ignore case, convert it to lower now so we don't have to
-	       * do it every time
-	       */
-	      lc_strcpy(pathpart, pathpart);
-	    }
+	    add_visitor(visit_substring_match_casefold, pathpart);
 	  else
-	    {
-	      add_visitor(visit_substring_match_nocasefold, pathpart);
-	    }
+	    add_visitor(visit_substring_match_nocasefold, pathpart);
 	}
     }
 
@@ -1301,14 +1283,15 @@ Usage: %s [-d path | --database=path] [-e | -E | --[non-]existing]\n\
       [--limit=N | -l N] [-S | --statistics] [-0 | --null] [-c | --count]\n\
       [-P | -H | --nofollow] [-L | --follow] [-m | --mmap ] [ -s | --stdio ]\n\
       [-A | --all] [-p | --print] [-r | --regex ] [--regextype=TYPE]\n\
-      [-version] [--help]\n\
+      [--max-database-age D] [--version] [--help]\n\
       pattern...\n"),
 	   program_name);
   fputs (_("\nReport bugs to <bug-findutils@gnu.org>.\n"), stream);
 }
 enum
   {
-    REGEXTYPE_OPTION = CHAR_MAX + 1
+    REGEXTYPE_OPTION = CHAR_MAX + 1,
+    MAX_DB_AGE
   };
 
 
@@ -1335,6 +1318,7 @@ static struct option const longopts[] =
   {"statistics",  no_argument, NULL, 'S'},
   {"follow",      no_argument, NULL, 'L'},
   {"nofollow",    no_argument, NULL, 'P'},
+  {"max-database-age",    required_argument, NULL, MAX_DB_AGE},
   {NULL, no_argument, NULL, 0}
 };
 
@@ -1343,11 +1327,14 @@ static int
 drop_privs(void)
 {
   const char * what = "failed";
-  uid_t orig_euid = geteuid();
-
+  const uid_t orig_euid = geteuid();
+  const uid_t uid       = getuid();
+  const gid_t gid       = getgid();
+  
   /* Use of setgroups() is restrcted to root only. */
   if (0 == orig_euid)
     {
+      /* UID != 0, but EUID == 0.  We're running setuid-root. */
       gid_t groups[1];
       groups[1] = getgid();
       if (0 != setgroups(1, groups)) 
@@ -1356,21 +1343,52 @@ drop_privs(void)
 	  goto fail;
 	}
     }
-  
-  if (0 != setuid(getuid()))
+
+  /* Drop any setuid privileges */
+  if (uid != orig_euid)
     {
-      what = _("failed to drop setuid privileges");
-      goto fail;
+      if (0 == uid)
+	{
+	  /* We're really root anyway, but are setuid to something else. Leave it. */
+	}
+      else
+	{
+	  errno = 0;
+	  if (0 != setuid(getuid()))
+	    {
+	      what = _("failed to drop setuid privileges");
+	      goto fail;
+	    }
+	  
+	  /* Defend against the case where the attacker runs us with the
+	   * capability to call setuid() turned off, which on some systems
+	   * will cause the above attempt to drop privileges fail (leaving us
+	   * privileged).
+	   */
+	  else
+	    {
+	      /* Check that we can no longer switch bask to root */
+	      if (0 == setuid(0))
+		{
+		  what = _("Failed to fully drop privileges");
+		  /* The errno value here is not interesting (since
+		   * the system call we are complaining about
+		   * succeeded when we wanted it to fail).  Arrange
+		   * for the call to error() not to print the errno
+		   * value by setting errno=0.
+		   */
+		  errno = 0;
+		  goto fail;
+		}
+	    }
+	}
     }
   
-  /* Defend against the case where the attacker runs us with the
-   * capability to call setuid() turned off, which on some systems
-   * will cause the above attempt to drop privileges fail (leaving us
-   * privileged).
-   */
-  if (0 == setuid(0))
+  /* Drop any setgid privileges */
+  errno = 0;
+  if (0 != setgid(gid))
     {
-      what = _("Failed to drop privileges");
+      what = _("failed to drop setgid privileges");
       goto fail;
     }
 
@@ -1498,12 +1516,18 @@ dolocate (int argc, char **argv, int secure_db_fd)
 	usage (stdout);
 	return 0;
 
+      case MAX_DB_AGE:
+	/* XXX: nothing in the test suite for this option. */
+	set_max_db_age(optarg);
+	break;
+
       case 'p':
 	print = 1;
 	break;
 
       case 'v':
 	printf (_("GNU locate version %s\n"), version_string);
+	printf (_("Built using GNU gnulib version %s\n"), gnulib_version);
 	return 0;
 
       case 'w':
@@ -1689,13 +1713,17 @@ dolocate (int argc, char **argv, int secure_db_fd)
 	    }
 	  else
 	    {
-	      if (now - st.st_mtime > WARN_SECONDS)
+	      double age          = difftime(now, st.st_mtime);
+	      double warn_seconds = SECONDS_PER_UNIT * warn_number_units;
+	      if (age > warn_seconds)
 		{
 		  /* For example:
-		     warning: database `fred' is more than 8 days old */
+		     warning: database `fred' is more than 8 days old (actual age is 10 days)*/
 		  error (0, 0,
-			 _("warning: database `%s' is more than %d %s old"),
-			 e, WARN_NUMBER_UNITS, _(warn_name_units));
+			 _("warning: database `%s' is more than %d %s old (actual age is %.1f %s)"),
+			 e, 
+			 warn_number_units,              _(warn_name_units),
+			 (age/(double)SECONDS_PER_UNIT), _(warn_name_units));
 		}
 	    }
 	}
